@@ -1,7 +1,8 @@
-__version__ = '0.5.2'
+__version__ = '0.5.4'
+
 
 # amalgamate exclude jupyter_kernel parser_table parser_test_table pyghooks
-# amalgamate exclude winutils wizard pytest_plugin fs
+# amalgamate exclude winutils wizard pytest_plugin fs macutils
 import os as _os
 if _os.getenv('XONSH_DEBUG', ''):
     pass
@@ -11,6 +12,8 @@ else:
         from xonsh import __amalgam__
         completer = __amalgam__
         _sys.modules['xonsh.completer'] = __amalgam__
+        contexts = __amalgam__
+        _sys.modules['xonsh.contexts'] = __amalgam__
         lazyasd = __amalgam__
         _sys.modules['xonsh.lazyasd'] = __amalgam__
         lazyjson = __amalgam__
@@ -23,14 +26,14 @@ else:
         _sys.modules['xonsh.platform'] = __amalgam__
         pretty = __amalgam__
         _sys.modules['xonsh.pretty'] = __amalgam__
-        timings = __amalgam__
-        _sys.modules['xonsh.timings'] = __amalgam__
         jobs = __amalgam__
         _sys.modules['xonsh.jobs'] = __amalgam__
         lazyimps = __amalgam__
         _sys.modules['xonsh.lazyimps'] = __amalgam__
         parser = __amalgam__
         _sys.modules['xonsh.parser'] = __amalgam__
+        style_tools = __amalgam__
+        _sys.modules['xonsh.style_tools'] = __amalgam__
         tokenize = __amalgam__
         _sys.modules['xonsh.tokenize'] = __amalgam__
         tools = __amalgam__
@@ -39,8 +42,6 @@ else:
         _sys.modules['xonsh.ast'] = __amalgam__
         commands_cache = __amalgam__
         _sys.modules['xonsh.commands_cache'] = __amalgam__
-        contexts = __amalgam__
-        _sys.modules['xonsh.contexts'] = __amalgam__
         diff_history = __amalgam__
         _sys.modules['xonsh.diff_history'] = __amalgam__
         events = __amalgam__
@@ -59,6 +60,8 @@ else:
         _sys.modules['xonsh.dirstack'] = __amalgam__
         inspectors = __amalgam__
         _sys.modules['xonsh.inspectors'] = __amalgam__
+        timings = __amalgam__
+        _sys.modules['xonsh.timings'] = __amalgam__
         xonfig = __amalgam__
         _sys.modules['xonsh.xonfig'] = __amalgam__
         environ = __amalgam__
