@@ -255,14 +255,23 @@ def source_alias(args, stdin=None):
     env = builtins.__xonsh_env__
     encoding = env.get('XONSH_ENCODING')
     errors = env.get('XONSH_ENCODING_ERRORS')
-    for fname in args:
-        if not os.path.isfile(fname):
-            fname = locate_binary(fname)
-        with open(fname, 'r', encoding=encoding, errors=errors) as fp:
+    for i, fname in enumerate(args):
+        fpath = fname
+        if not os.path.isfile(fpath):
+            fpath = locate_binary(fname)
+            if fpath is None:
+                if env.get('XONSH_DEBUG'):
+                    print('source: {}: No such file'.format(fname), file=sys.stderr)
+                if i == 0:
+                    raise RuntimeError('must source at least one file, ' + fname +
+                                       'does not exist.')
+                break
+        with open(fpath, 'r', encoding=encoding, errors=errors) as fp:
             src = fp.read()
         if not src.endswith('\n'):
             src += '\n'
-        builtins.execx(src, 'exec', builtins.__xonsh_ctx__)
+        with env.swap(ARGS=args[i+1:]):
+            builtins.execx(src, 'exec', builtins.__xonsh_ctx__, filename=fpath)
 
 
 def source_cmd(args, stdin=None):
@@ -371,6 +380,27 @@ def showcmd(args, stdin=None):
         sys.displayhook(args)
 
 
+def detect_xip_alias():
+    """
+    Determines the correct invokation to get xonsh's pip
+    """
+    if not getattr(sys, 'executable', None):
+        return lambda args, stdin=None: ("", "Sorry, unable to run pip on your system (missing sys.executable)", 1)
+
+    basecmd = [sys.executable, '-m', 'pip']
+    try:
+        if ON_WINDOWS:
+            # XXX: Does windows have an installation mode that requires UAC?
+            return basecmd
+        elif not os.access(os.path.dirname(sys.executable), os.W_OK):
+            return ['sudo'] + basecmd
+        else:
+            return basecmd
+    except Exception:
+        # Something freaky happened, return something that'll probably work
+        return basecmd
+
+
 def make_default_aliases():
     """Creates a new default aliases dictionary."""
     default_aliases = {
@@ -401,7 +431,8 @@ def make_default_aliases():
         'ipynb': ['jupyter', 'notebook', '--no-browser'],
         'which': xxw.which,
         'xontrib': xontribs_main,
-        'completer': xca.completer_alias
+        'completer': xca.completer_alias,
+        'xip': detect_xip_alias(),
     }
     if ON_WINDOWS:
         # Borrow builtin commands from cmd.exe.
