@@ -15,7 +15,7 @@ from xonsh.codecache import (should_use_cache, code_cache_name,
 from xonsh.completer import Completer
 from xonsh.prompt.base import multiline_prompt, PromptFormatter
 from xonsh.events import events
-from xonsh.shell import fire_precommand
+from xonsh.shell import transform_command
 
 if ON_WINDOWS:
     import ctypes
@@ -308,6 +308,9 @@ class BaseShell(object):
         src, code = self.push(line)
         if code is None:
             return
+
+        events.on_precommand.fire(cmd=src)
+
         env = builtins.__xonsh_env__
         hist = builtins.__xonsh_history__  # pylint: disable=no-member
         ts1 = None
@@ -355,10 +358,10 @@ class BaseShell(object):
         else:
             info['out'] = tee_out + '\n' + last_out
         events.on_postcommand.fire(
-            info['inp'],
-            info['rtn'],
-            info.get('out', None),
-            info['ts']
+            cmd=info['inp'],
+            rtn=info['rtn'],
+            out=info.get('out', None),
+            ts=info['ts']
             )
         if hist is not None:
             hist.append(info)
@@ -372,7 +375,7 @@ class BaseShell(object):
             builtins.__xonsh_env__['PWD'] = cwd      # track it now
             if old is not None:
                 builtins.__xonsh_env__['OLDPWD'] = old  # and update $OLDPWD like dirstack.
-            events.on_chdir.fire(old, cwd)              # fire event after cwd actually changed.
+            events.on_chdir.fire(olddir=old, newdir=cwd)              # fire event after cwd actually changed.
 
     def push(self, line):
         """Pushes a line onto the buffer and compiles the code in a way that
@@ -382,7 +385,7 @@ class BaseShell(object):
         if self.need_more_lines:
             return None, None
         src = ''.join(self.buffer)
-        src = fire_precommand(src)
+        src = transform_command(src)
         return self.compile(src)
 
     def compile(self, src):
@@ -490,3 +493,10 @@ class BaseShell(object):
     def color_style(self):
         """Returns the current color map."""
         return {}
+
+    def restore_tty_sanity(self):
+        """An interface for resetting the TTY stdin mode. This is highly
+        dependent on the shell backend. Also it is mostly optional since
+        it only affects ^Z backgrounding behaviour.
+        """
+        pass
