@@ -34,10 +34,10 @@ from xonsh.proc import (
     pause_call_resume, CommandPipeline, HiddenCommandPipeline,
     STDOUT_CAPTURE_KINDS)
 from xonsh.tools import (
-    suggest_commands, expandvars, globpath, XonshError,
-    XonshCalledProcessError, XonshBlockError
+    suggest_commands, expand_path, globpath, XonshError,
+    XonshCalledProcessError
 )
-from xonsh.lazyimps import pty
+from xonsh.lazyimps import pty, termios
 from xonsh.commands_cache import CommandsCache
 from xonsh.events import events
 
@@ -82,23 +82,6 @@ def superhelper(x, name=''):
     return x
 
 
-def expand_path(s):
-    """Takes a string path and expands ~ to home and environment vars."""
-    if builtins.__xonsh_env__.get('EXPAND_ENV_VARS'):
-        s = expandvars(s)
-    # expand ~ according to Bash unquoted rules "Each variable assignment is
-    # checked for unquoted tilde-prefixes immediately following a ':' or the
-    # first '='". See the following for more details.
-    # https://www.gnu.org/software/bash/manual/html_node/Tilde-Expansion.html
-    pre, char, post = s.partition('=')
-    if char:
-        s = os.path.expanduser(pre) + char
-        s += os.pathsep.join(map(os.path.expanduser, post.split(os.pathsep)))
-    else:
-        s = os.path.expanduser(s)
-    return s
-
-
 def reglob(path, parts=None, i=None):
     """Regular expression-based globbing."""
     if parts is None:
@@ -265,6 +248,10 @@ _E2O_MAP = LazyObject(lambda: frozenset({'{}>{}'.format(e, o)
                                          for e in _REDIR_ERR
                                          for o in _REDIR_OUT
                                          if o != ''}), globals(), '_E2O_MAP')
+_O2E_MAP = LazyObject(lambda: frozenset({'{}>{}'.format(o, e)
+                                         for e in _REDIR_ERR
+                                         for o in _REDIR_OUT
+                                         if o != ''}), globals(), '_O2E_MAP')
 
 
 def _is_redirect(x):
@@ -296,7 +283,7 @@ def safe_close(x):
         pass
 
 
-def _parse_redirects(r):
+def _parse_redirects(r, loc=None):
     """returns origin, mode, destination tuple"""
     orig, mode, dest = _REDIR_REGEX.match(r).groups()
     # redirect to fd
@@ -323,10 +310,14 @@ def _parse_redirects(r):
 def _redirect_streams(r, loc=None):
     """Returns stdin, stdout, stderr tuple of redirections."""
     stdin = stdout = stderr = None
+    no_ampersand = r.replace('&', '')
     # special case of redirecting stderr to stdout
-    if r.replace('&', '') in _E2O_MAP:
+    if no_ampersand in _E2O_MAP:
         stderr = subprocess.STDOUT
         return stdin, stdout, stderr
+    elif no_ampersand in _O2E_MAP:
+        stdout = 2  # using 2 as a flag, rather than using a file object
+        return stdin, stdout, stderr
     # get streams
     orig, mode, dest = _parse_redirects(r)
     if mode == 'r':
@@ -660,6 +651,20 @@ class SubprocSpec:
         self.captured = cpable
 
 
+def _safe_pipe_properties(fd, use_tty=False):
+    """Makes sure that a pipe file descriptor properties are sane."""
+    if not use_tty:
+        return
+    # due to some weird, long standing issue in Python, PTYs come out
+    # replacing newline \n with \r\n. This causes issues for raw unix
+    # protocols, like git and ssh, which expect unix line endings.
+    # see https://mail.python.org/pipermail/python-list/2013-June/650460.html
+    # for more details and the following solution.
+    props = termios.tcgetattr(fd)
+    props[1] = props[1] & (~termios.ONLCR) | termios.ONLRET
+    termios.tcsetattr(fd, termios.TCSANOW, props)
+
+
 def _update_last_spec(last):
     captured = last.captured
     last.last_in_pipeline = True
@@ -699,7 +704,9 @@ def _update_last_spec(last):
     else:
         last.universal_newlines = True
         r, w = pty.openpty() if use_tty else os.pipe()
+        _safe_pipe_properties(w, use_tty=use_tty)
         last.stdout = safe_open(w, 'w')
+        _safe_pipe_properties(r, use_tty=use_tty)
         last.captured_stdout = safe_open(r, 'r')
     # set standard error
     if last.stderr is not None:
@@ -716,8 +723,14 @@ def _update_last_spec(last):
         last.stderr = None  # must truly stream on windows
     else:
         r, w = pty.openpty() if use_tty else os.pipe()
+        _safe_pipe_properties(w, use_tty=use_tty)
         last.stderr = safe_open(w, 'w')
+        _safe_pipe_properties(r, use_tty=use_tty)
         last.captured_stderr = safe_open(r, 'r')
+    # redirect stdout to stderr, if we should
+    if isinstance(last.stdout, int) and last.stdout == 2:
+        # need to use private interface to avoid duplication.
+        last._stdout = last.stderr
 
 
 def cmds_to_specs(cmds, captured=False):
@@ -892,7 +905,7 @@ def MACRO_FLAG_KINDS():
         'exec': exec,
         't': type,
         'type': type,
-        }
+    }
 
 
 def _convert_kind_flag(x):
@@ -1147,7 +1160,6 @@ def load_builtins(execer=None, config=None, login=False, ctx=None):
     builtins.__xonsh_path_literal__ = path_literal
     # public built-ins
     builtins.XonshError = XonshError
-    builtins.XonshBlockError = XonshBlockError
     builtins.XonshCalledProcessError = XonshCalledProcessError
     builtins.evalx = None if execer is None else execer.eval
     builtins.execx = None if execer is None else execer.exec
@@ -1214,7 +1226,6 @@ def unload_builtins():
              '__xonsh_enter_macro__',
              '__xonsh_path_literal__',
              'XonshError',
-             'XonshBlockError',
              'XonshCalledProcessError',
              'evalx',
              'execx',
