@@ -7,12 +7,12 @@ and returns whethere or not the process can be run in the background (returns
 True) or must be run the foreground (returns False).
 """
 import os
+import time
 import builtins
 import argparse
-import collections
 import collections.abc as cabc
 
-from xonsh.platform import ON_WINDOWS, pathbasename
+from xonsh.platform import ON_WINDOWS, ON_POSIX, pathbasename
 from xonsh.tools import executables_in
 from xonsh.lazyasd import lazyobject
 
@@ -79,7 +79,7 @@ class CommandsCache(cabc.Mapping):
         cache_valid = path_hash == self._path_checksum
         self._path_checksum = path_hash
         # did aliases change?
-        alss = getattr(builtins, 'aliases', set())
+        alss = getattr(builtins, 'aliases', dict())
         al_hash = hash(frozenset(alss))
         cache_valid = cache_valid and al_hash == self._alias_checksum
         self._alias_checksum = al_hash
@@ -99,7 +99,7 @@ class CommandsCache(cabc.Mapping):
             # entries at the back.
             for cmd in executables_in(path):
                 key = cmd.upper() if ON_WINDOWS else cmd
-                allcmds[key] = (os.path.join(path, cmd), cmd in alss)
+                allcmds[key] = (os.path.join(path, cmd), alss.get(key, None))
         for cmd in alss:
             if cmd not in allcmds:
                 key = cmd.upper() if ON_WINDOWS else cmd
@@ -142,14 +142,32 @@ class CommandsCache(cabc.Mapping):
         """A lazy value getter."""
         return self._cmds_cache.get(self.cached_name(key), default)
 
-    def locate_binary(self, name):
-        """Locates an executable on the file system using the cache."""
+    def locate_binary(self, name, ignore_alias=False):
+        """Locates an executable on the file system using the cache.
+
+        Arguments
+        ---------
+        name : str
+                name of binary to search for
+        ignore_alias : bool, optional
+                Force return of binary path even if alias of ``name`` exists
+                (default ``False``)
+        """
         # make sure the cache is up to date by accessing the property
         _ = self.all_commands
-        return self.lazy_locate_binary(name)
+        return self.lazy_locate_binary(name, ignore_alias)
 
-    def lazy_locate_binary(self, name):
-        """Locates an executable in the cache, without checking its validity."""
+    def lazy_locate_binary(self, name, ignore_alias=False):
+        """Locates an executable in the cache, without checking its validity.
+
+        Arguments
+        ---------
+        name : str
+                name of binary to search for
+        ignore_alias : bool, optional
+                Force return of binary path even if alias of ``name`` exists
+                (default ``False``)
+        """
         possibilities = self.get_possible_names(name)
         if ON_WINDOWS:
             # Windows users expect to be able to execute files in the same
@@ -161,8 +179,11 @@ class CommandsCache(cabc.Mapping):
         cached = next((cmd for cmd in possibilities if cmd in self._cmds_cache),
                       None)
         if cached:
-            (path, is_alias) = self._cmds_cache[cached]
-            return path if not is_alias else None
+            (path, alias) = self._cmds_cache[cached]
+            if not alias or ignore_alias:
+                return path
+            else:
+                return None
         elif os.path.isfile(name) and name != pathbasename(name):
             return name
 
@@ -184,9 +205,74 @@ class CommandsCache(cabc.Mapping):
                 pre, ext = os.path.splitext(name)
                 if pre in predictors:
                     predictors[name] = predictors[pre]
+        if name not in predictors:
+            predictors[name] = self.default_predictor(name, cmd[0])
         predictor = predictors[name]
         return predictor(cmd[1:])
 
+    #
+    # Background Predictors (as methods)
+    #
+
+    def default_predictor(self, name, cmd0):
+        if ON_POSIX:
+            return self.default_predictor_readbin(name, cmd0,
+                                                  timeout=.1,
+                                                  failure=predict_true)
+        else:
+            return predict_true
+
+    def default_predictor_readbin(self, name, cmd0, timeout, failure):
+        """Make a defautt predictor by
+        analyzing the content of the binary. Should only works on POSIX.
+        Return failure if the analysis fails.
+        """
+        fname = cmd0 if os.path.isabs(cmd0) else None
+        fname = cmd0 if fname is None and os.sep in cmd0 else fname
+        fname = self.lazy_locate_binary(name) if fname is None else fname
+
+        if fname is None:
+            return failure
+        if not os.path.isfile(fname):
+            return failure
+
+        try:
+            fd = os.open(fname, os.O_RDONLY | os.O_NONBLOCK)
+        except Exception:
+            return failure  # opening error
+
+        search_for = {
+            (b'ncurses',): [False, ],
+            (b'isatty', b'tcgetattr', b'tcsetattr'): [False, False, False],
+        }
+        tstart = time.time()
+        block = b''
+        while time.time() < tstart + timeout:
+            previous_block = block
+            try:
+                block = os.read(fd, 2048)
+            except Exception:
+                # should not occur, except e.g. if a file is deleted a a dir is
+                # created with the same name between os.path.isfile and os.open
+                os.close(fd)
+                return failure
+            if len(block) == 0:
+                os.close(fd)
+                return predict_true  # no keys of search_for found
+            analyzed_block = previous_block + block
+            for k, v in search_for.items():
+                for i in range(len(k)):
+                    if v[i]:
+                        continue
+                    if k[i] in analyzed_block:
+                        v[i] = True
+                if all(v):
+                    os.close(fd)
+                    return predict_false  # use one key of search_for
+        os.close(fd)
+        return failure  # timeout
+
+
 #
 # Background Predictors
 #
@@ -258,6 +344,7 @@ def default_threadable_predictors():
         'cmd': predict_shell,
         'ex': predict_false,
         'fish': predict_shell,
+        'gvim': predict_help_ver,
         'htop': predict_help_ver,
         'ksh': predict_shell,
         'less': predict_help_ver,
@@ -267,6 +354,9 @@ def default_threadable_predictors():
         'mutt': predict_help_ver,
         'nano': predict_help_ver,
         'psql': predict_false,
+        'python': predict_shell,
+        'python2': predict_shell,
+        'python3': predict_shell,
         'ranger': predict_help_ver,
         'rview': predict_false,
         'rvim': predict_false,
@@ -276,6 +366,7 @@ def default_threadable_predictors():
         'startx': predict_false,
         'sudo': predict_help_ver,
         'tcsh': predict_shell,
+        'telnet': predict_false,
         'top': predict_help_ver,
         'vi': predict_false,
         'view': predict_false,
@@ -286,4 +377,4 @@ def default_threadable_predictors():
         'xonsh': predict_shell,
         'zsh': predict_shell,
     }
-    return collections.defaultdict(lambda: predict_true, predictors)
+    return predictors
