@@ -18,19 +18,32 @@ def RE_ATTR():
 
 
 @xl.lazyobject
-def XONSH_TOKENS():
+def XONSH_EXPR_TOKENS():
     return {
-        'and ', 'as ', 'assert ', 'break', 'class ', 'continue', 'def ', 'del ',
-        'elif ', 'else', 'except ', 'finally:', 'for ', 'from ', 'global ',
-        'import ', 'if ', 'in ', 'is ', 'lambda ', 'nonlocal ', 'not ', 'or ',
-        'pass', 'raise ', 'return ', 'try:', 'while ', 'with ', 'yield ', '+',
-        '-', '/', '//', '%', '**', '|', '&', '~', '^', '>>', '<<', '<', '<=',
-        '>', '>=', '==', '!=', '->', '=', '+=', '-=', '*=', '/=', '%=', '**=',
-        '>>=', '<<=', '&=', '^=', '|=', '//=', ',', ';', ':', '?', '??', '$(',
-        '${', '$[', '..', '...', '![', '!(', '@(', '@$(', '@'
+        'and ', 'else', 'for ', 'if ', 'in ', 'is ', 'lambda ', 'not ', 'or ',
+        '+', '-', '/', '//', '%', '**', '|', '&', '~', '^', '>>', '<<', '<',
+        '<=', '>', '>=', '==', '!=', ',', '?', '??', '$(',
+        '${', '$[', '...', '![', '!(', '@(', '@$(', '@',
         }
 
 
+@xl.lazyobject
+def XONSH_STMT_TOKENS():
+    return {
+        'as ', 'assert ', 'break', 'class ', 'continue', 'def ', 'del ',
+        'elif ', 'except ', 'finally:', 'from ', 'global ', 'import ',
+        'nonlocal ', 'pass', 'raise ', 'return ', 'try:', 'while ', 'with ',
+        'yield ', '-', '/', '//', '%', '**', '|', '&', '~', '^', '>>', '<<',
+        '<', '<=', '->', '=', '+=', '-=', '*=', '/=', '%=', '**=',
+        '>>=', '<<=', '&=', '^=', '|=', '//=', ';', ':', '..',
+        }
+
+
+@xl.lazyobject
+def XONSH_TOKENS():
+    return set(XONSH_EXPR_TOKENS) | set(XONSH_STMT_TOKENS)
+
+
 def complete_python(prefix, line, start, end, ctx):
     """
     Completes based on the contents of the current Python environment,
@@ -57,11 +70,20 @@ def _complete_python(prefix, line, start, end, ctx):
         if first in builtins.__xonsh_commands_cache__ and first not in ctx:
             return set()
     filt = get_filter_function()
-    rtn = {s for s in XONSH_TOKENS if filt(s, prefix)}
+    rtn = set()
     if ctx is not None:
         if '.' in prefix:
             rtn |= attr_complete(prefix, ctx, filt)
+        args = python_signature_complete(prefix, line, end, ctx, filt)
+        rtn |= args
         rtn |= {s for s in ctx if filt(s, prefix)}
+    else:
+        args = ()
+    if len(args) == 0:
+        # not in a function call, so we can add non-expression tokens
+        rtn |= {s for s in XONSH_TOKENS if filt(s, prefix)}
+    else:
+        rtn |= {s for s in XONSH_EXPR_TOKENS if filt(s, prefix)}
     rtn |= {s for s in dir(builtins) if filt(s, prefix)}
     return rtn
 
@@ -79,6 +101,24 @@ def complete_python_mode(prefix, line, start, end, ctx):
     return set(prefix_start + i for i in python_matches)
 
 
+def _safe_eval(expr, ctx):
+    """Safely tries to evaluate an expression. If this fails, it will return
+    a (None, None) tuple.
+    """
+    _ctx = None
+    xonsh_safe_eval = builtins.__xonsh_execer__.eval
+    try:
+        val = xonsh_safe_eval(expr, ctx, ctx, transform=False)
+        _ctx = ctx
+    except:  # pylint:disable=bare-except
+        try:
+            val = xonsh_safe_eval(expr, builtins.__dict__, transform=False)
+            _ctx = builtins.__dict__
+        except:  # pylint:disable=bare-except
+            val = _ctx = None
+    return val, _ctx
+
+
 def attr_complete(prefix, ctx, filter_func):
     """Complete attributes of an object."""
     attrs = set()
@@ -89,17 +129,9 @@ def attr_complete(prefix, ctx, filter_func):
     expr = xt.subexpr_from_unbalanced(expr, '(', ')')
     expr = xt.subexpr_from_unbalanced(expr, '[', ']')
     expr = xt.subexpr_from_unbalanced(expr, '{', '}')
-    _ctx = None
-    xonsh_safe_eval = builtins.__xonsh_execer__.eval
-    try:
-        val = xonsh_safe_eval(expr, ctx, transform=False)
-        _ctx = ctx
-    except:  # pylint:disable=bare-except
-        try:
-            val = xonsh_safe_eval(expr, builtins.__dict__, transform=False)
-            _ctx = builtins.__dict__
-        except:  # pylint:disable=bare-except
-            return attrs  # anything could have gone wrong!
+    val, _ctx = _safe_eval(expr, ctx)
+    if val is None and _ctx is None:
+        return attrs
     if len(attr) == 0:
         opts = [o for o in dir(val) if not o.startswith('_')]
     else:
@@ -107,10 +139,9 @@ def attr_complete(prefix, ctx, filter_func):
     prelen = len(prefix)
     for opt in opts:
         # check whether these options actually work (e.g., disallow 7.imag)
-        try:
-            _val = '{0}.{1}'.format(expr, opt)
-            xonsh_safe_eval(_val, _ctx, transform=False)
-        except:  # pylint:disable=bare-except
+        _expr = '{0}.{1}'.format(expr, opt)
+        _val_, _ctx_ = _safe_eval(_expr, _ctx)
+        if _val_ is None and _ctx_ is None:
             continue
         a = getattr(val, opt)
         if builtins.__xonsh_env__['COMPLETIONS_BRACKETS']:
@@ -129,6 +160,22 @@ def attr_complete(prefix, ctx, filter_func):
     return attrs
 
 
+def python_signature_complete(prefix, line, end, ctx, filter_func):
+    """Completes a python function (or other callable) call by completing
+    argument and keyword argument names.
+    """
+    front = line[:end]
+    if xt.is_balanced(front, '(', ')'):
+        return set()
+    funcname = xt.subexpr_before_unbalanced(front, '(', ')')
+    val, _ctx = _safe_eval(funcname, ctx)
+    if val is None and _ctx is None:
+        return set()
+    sig = inspect.signature(val)
+    args = {p + '=' for p in sig.parameters if filter_func(p, prefix)}
+    return args
+
+
 def complete_import(prefix, line, start, end, ctx):
     """
     Completes module names and contents for "import ..." and "from ... import
