@@ -56,8 +56,26 @@ class _LazyModule(_ModuleType):
 #
 # -*- coding: utf-8 -*-
 """Base class of Xonsh History backends."""
+types = _LazyModule.load('types', 'types')
 uuid = _LazyModule.load('uuid', 'uuid')
-xt = _LazyModule.load('xonsh', 'xonsh.tools', 'xt')
+class HistoryEntry(types.SimpleNamespace):
+    """Represent a command in history.
+
+    Attributes
+    ----------
+    cmd: str
+        The command as typed by the user, including newlines
+    out: str
+        The output of the command, if xonsh is configured to save it
+    rtn: int
+        The return of the command (ie, 0 on success)
+    ts: two-tuple of floats
+        The timestamps of when the command started and finished, including
+        fractions.
+
+    """
+
+
 class History:
     """Xonsh history backend base class.
 
@@ -65,31 +83,10 @@ class History:
 
     Indexing
     --------
-    History object acts like a sequence that can be indexed in a special
-    way that adds extra functionality. Note that the most recent command
-    is the last item in history.
+    History acts like a sequence that can be indexed to return
+    ``HistoryEntry`` objects.
 
-    The index acts as a filter with two parts, command and argument,
-    separated by comma. Based on the type of each part different
-    filtering can be achieved,
-
-        for the command part:
-
-            - an int returns the command in that position.
-            - a slice returns a list of commands.
-            - a string returns the most recent command containing the string.
-
-        for the argument part:
-
-            - an int returns the argument of the command in that position.
-            - a slice returns a part of the command based on the argument
-              position.
-
-    The argument part of the filter can be omitted but the command part is
-    required. Command arguments are separated by white space.
-
-    If the filtering produces only one result it is returned as a string
-    else a list of strings is returned.
+    Note that the most recent command is the last item in history.
 
     Attributes
     ----------
@@ -133,23 +130,22 @@ class History:
         return len(list(self.items()))
 
     def __getitem__(self, item):
-        """Retrieve history parts based on filtering rules,
-        see ``History`` docs for more info. Accepts one of
-        int, string, slice or tuple of length two.
-        """
-        if isinstance(item, tuple):
-            cmd_pat, arg_pat = item
+        """Retrieve history entries, see ``History`` docs for more info."""
+        if isinstance(item, int):
+            if item >= len(self):
+                raise IndexError('history index out of range')
+            return HistoryEntry(cmd=self.inps[item], out=self.outs[item],
+                                rtn=self.rtns[item], ts=self.tss[item])
+        elif isinstance(item, slice):
+            cmds = self.inps[item]
+            outs = self.outs[item]
+            rtns = self.rtns[item]
+            tss = self.tss[item]
+            return [HistoryEntry(cmd=c, out=o, rtn=r, ts=t)
+                    for c, o, r, t in zip(cmds, outs, rtns, tss)]
         else:
-            cmd_pat, arg_pat = item, None
-        cmds = [c['inp'] for c in self.items()]
-        cmds = self._cmd_filter(cmds, cmd_pat)
-        if arg_pat is not None:
-            cmds = self._args_filter(cmds, arg_pat)
-        cmds = list(cmds)
-        if len(cmds) == 1:
-            return cmds[0]
-        else:
-            return cmds
+            raise TypeError('history indices must be integers '
+                            'or slices, not {}'.format(type(item)))
 
     def __setitem__(self, *args):
         raise PermissionError('You cannot change history! '
@@ -200,30 +196,6 @@ class History:
         """
         pass
 
-    @staticmethod
-    def _cmd_filter(cmds, pat):
-        if isinstance(pat, (int, slice)):
-            s = xt.ensure_slice(pat)
-            yield from xt.get_portions(cmds, s)
-        elif xt.is_string(pat):
-            for command in reversed(list(cmds)):
-                if pat in command:
-                    yield command
-                    return
-        else:
-            raise TypeError('Command filter must be string, int or slice')
-
-    @staticmethod
-    def _args_filter(cmds, pat):
-        args = None
-        if isinstance(pat, (int, slice)):
-            s = xt.ensure_slice(pat)
-            for command in cmds:
-                yield ' '.join(command.split()[s])
-        else:
-            raise TypeError('Argument filter must be int or slice')
-        return args
-
 #
 # dummy
 #
@@ -260,8 +232,9 @@ builtins = _LazyModule.load('builtins', 'builtins')
 threading = _LazyModule.load('threading', 'threading')
 cabc = _LazyModule.load('collections', 'collections.abc', 'cabc')
 # amalgamated xonsh.history.base
-# amalgamated xonsh.tools
+xt = _LazyModule.load('xonsh', 'xonsh.tools', 'xt')
 xlj = _LazyModule.load('xonsh', 'xonsh.lazyjson', 'xlj')
+uptime = _LazyModule.load('xonsh', 'xonsh.xoreutils.uptime', 'uptime')
 def _xhj_gc_commands_to_rmfiles(hsize, files):
     """Return the history files to remove to get under the command limit."""
     rmfiles = []
@@ -372,7 +345,7 @@ class JsonHistoryGC(threading.Thread):
         env = getattr(builtins, '__xonsh_env__', None)
         if env is None:
             return []
-
+        boot = uptime.boottime()
         fs = _xhj_get_history_files(sort=False)
         files = []
         for f in fs:
@@ -382,10 +355,19 @@ class JsonHistoryGC(threading.Thread):
                     files.append((time.time(), 0, f))
                     continue
                 lj = xlj.LazyJSON(f, reopen=False)
+                if lj['locked'] and lj['ts'][0] < boot:
+                    # computer was rebooted between when this history was created
+                    # and now and so this history should be unlocked.
+                    hist = lj.load()
+                    lj.close()
+                    hist['locked'] = False
+                    with open(f, 'w', newline='\n') as fp:
+                        xlj.ljdump(hist, fp, sort_keys=True)
+                    lj = xlj.LazyJSON(f, reopen=False)
                 if only_unlocked and lj['locked']:
                     continue
                 # info: closing timestamp, number of commands, filename
-                files.append((lj['ts'][1] or time.time(),
+                files.append((lj['ts'][1] or lj['ts'][0],
                               len(lj.sizes['cmds']) - 1,
                               f))
                 lj.close()
@@ -903,6 +885,7 @@ functools = _LazyModule.load('functools', 'functools')
 # amalgamated json
 # amalgamated os
 # amalgamated sys
+# amalgamated xonsh.history.base
 # amalgamated xonsh.history.dummy
 # amalgamated xonsh.history.json
 # amalgamated xonsh.history.sqlite
@@ -920,12 +903,16 @@ def construct_history(**kwargs):
     """Construct the history backend object."""
     env = builtins.__xonsh_env__
     backend = env.get('XONSH_HISTORY_BACKEND')
-    if backend not in HISTORY_BACKENDS:
+    if isinstance(backend, str) and backend in HISTORY_BACKENDS:
+        kls_history = HISTORY_BACKENDS[backend]
+    elif xt.is_class(backend):
+        kls_history = backend
+    elif isinstance(backend, History):
+        return backend
+    else:
         print('Unknown history backend: {}. Using JSON version'.format(
             backend), file=sys.stderr)
         kls_history = JsonHistory
-    else:
-        kls_history = HISTORY_BACKENDS[backend]
     return kls_history(**kwargs)
 
 
