@@ -11,7 +11,8 @@ import subprocess
 import collections
 
 from xonsh.lazyasd import LazyObject
-from xonsh.platform import ON_DARWIN, ON_WINDOWS, ON_CYGWIN
+from xonsh.platform import ON_DARWIN, ON_WINDOWS, ON_CYGWIN, LIBC
+
 
 tasks = LazyObject(collections.deque, globals(), 'tasks')
 # Track time stamp of last exit command, so that two consecutive attempts to
@@ -69,7 +70,7 @@ if ON_WINDOWS:
     def _set_pgrp(info):
         pass
 
-    def wait_for_active_job(signal_to_send=None):
+    def wait_for_active_job(last_task=None, backgrounded=False):
         """
         Wait for the active job to finish, to be killed by SIGINT, or to be
         suspended by ctrl-z.
@@ -78,7 +79,7 @@ if ON_WINDOWS:
         active_task = get_next_task()
         # Return when there are no foreground active task
         if active_task is None:
-            return
+            return last_task
         obj = active_task['obj']
         _continue(active_task)
         while obj.returncode is None:
@@ -88,7 +89,7 @@ if ON_WINDOWS:
                 pass
             except KeyboardInterrupt:
                 _kill(active_task)
-        return wait_for_active_job()
+        return wait_for_active_job(last_task=active_task)
 
 else:
     def _continue(job):
@@ -133,9 +134,6 @@ else:
     #    give_terminal_to from bash 4.3 source, jobs.c, line 4030
     # this will give the terminal to the process group pgid
     if ON_CYGWIN:
-        _libc = LazyObject(lambda: ctypes.CDLL('cygwin1.dll'),
-                           globals(), '_libc')
-
         # on cygwin, signal.pthread_sigmask does not exist in Python, even
         # though pthread_sigmask is defined in the kernel.  thus, we use
         # ctypes to mimic the calls in the "normal" version below.
@@ -144,16 +142,16 @@ else:
             if st is not None and os.isatty(st):
                 omask = ctypes.c_ulong()
                 mask = ctypes.c_ulong()
-                _libc.sigemptyset(ctypes.byref(mask))
+                LIBC.sigemptyset(ctypes.byref(mask))
                 for i in _block_when_giving:
-                    _libc.sigaddset(ctypes.byref(mask), ctypes.c_int(i))
-                _libc.sigemptyset(ctypes.byref(omask))
-                _libc.sigprocmask(ctypes.c_int(signal.SIG_BLOCK),
-                                  ctypes.byref(mask),
-                                  ctypes.byref(omask))
-                _libc.tcsetpgrp(ctypes.c_int(st), ctypes.c_int(pgid))
-                _libc.sigprocmask(ctypes.c_int(signal.SIG_SETMASK),
-                                  ctypes.byref(omask), None)
+                    LIBC.sigaddset(ctypes.byref(mask), ctypes.c_int(i))
+                LIBC.sigemptyset(ctypes.byref(omask))
+                LIBC.sigprocmask(ctypes.c_int(signal.SIG_BLOCK),
+                                 ctypes.byref(mask),
+                                 ctypes.byref(omask))
+                LIBC.tcsetpgrp(ctypes.c_int(st), ctypes.c_int(pgid))
+                LIBC.sigprocmask(ctypes.c_int(signal.SIG_SETMASK),
+                                 ctypes.byref(omask), None)
     else:
         def _give_terminal_to(pgid):
             st = _shell_tty()
@@ -163,7 +161,7 @@ else:
                 os.tcsetpgrp(st, pgid)
                 signal.pthread_sigmask(signal.SIG_SETMASK, oldmask)
 
-    def wait_for_active_job():
+    def wait_for_active_job(last_task=None, backgrounded=False):
         """
         Wait for the active job to finish, to be killed by SIGINT, or to be
         suspended by ctrl-z.
@@ -173,17 +171,25 @@ else:
         # Return when there are no foreground active task
         if active_task is None:
             _give_terminal_to(_shell_pgrp)  # give terminal back to the shell
-            return
+            if backgrounded and hasattr(builtins, '__xonsh_shell__'):
+                # restoring sanity could probably be called whenever we return
+                # control to the shell. But it only seems to matter after a
+                # ^Z event. This *has* to be called after we give the terminal
+                # back to the shell.
+                builtins.__xonsh_shell__.shell.restore_tty_sanity()
+            return last_task
         pgrp = active_task.get('pgrp', None)
         obj = active_task['obj']
+        backgrounded = False
         # give the terminal over to the fg process
         if pgrp is not None:
             _give_terminal_to(pgrp)
         _continue(active_task)
         _, wcode = os.waitpid(obj.pid, os.WUNTRACED)
         if os.WIFSTOPPED(wcode):
-            print()  # get a newline because ^Z will have been printed
+            print('^Z')
             active_task['status'] = "stopped"
+            backgrounded = True
         elif os.WIFSIGNALED(wcode):
             print()  # get a newline because ^C will have been printed
             obj.signal = (os.WTERMSIG(wcode), os.WCOREDUMP(wcode))
@@ -191,7 +197,8 @@ else:
         else:
             obj.returncode = os.WEXITSTATUS(wcode)
             obj.signal = None
-        return wait_for_active_job()
+        return wait_for_active_job(last_task=active_task,
+                                   backgrounded=backgrounded)
 
 
 def get_next_task():
@@ -277,8 +284,8 @@ def clean_jobs():
         if builtins.__xonsh_all_jobs__:
             global _last_exit_time
             hist = builtins.__xonsh_history__
-            if hist is not None and hist.buffer:
-                last_cmd_start = builtins.__xonsh_history__.buffer[-1]['ts'][0]
+            if hist is not None and len(hist.tss) > 0:
+                last_cmd_start = hist.tss[-1][0]
             else:
                 last_cmd_start = None
 
@@ -342,7 +349,6 @@ def fg(args, stdin=None):
     given as an argument, bring that job to the foreground. Additionally,
     specify "+" for the most recent job and "-" for the second most recent job.
     """
-
     _clear_dead_jobs()
     if len(tasks) == 0:
         return '', 'Cannot bring nonexistent job to foreground.\n'
@@ -373,6 +379,7 @@ def fg(args, stdin=None):
     job['bg'] = False
     job['status'] = "running"
     print_one_job(act)
+    wait_for_active_job()
 
 
 def bg(args, stdin=None):
