@@ -9,6 +9,7 @@ import contextlib
 import traceback
 
 from xonsh import __version__
+from xonsh.timings import setup_timings
 from xonsh.lazyasd import lazyobject
 from xonsh.shell import Shell
 from xonsh.pretty import pretty
@@ -157,6 +158,13 @@ def parser():
                    dest='shell_type',
                    choices=('readline', 'prompt_toolkit', 'best', 'random'),
                    default=None)
+    p.add_argument('--timings',
+                   help='Prints timing infomation before the prompt is shown. '
+                        'This is usefull to track down perfomance issues '
+                        'and investigate startup times.',
+                   dest='timings',
+                   action='store_true',
+                   default=None)
     p.add_argument('file',
                    metavar='script-file',
                    help='If present, execute the script in script-file'
@@ -203,6 +211,7 @@ def premain(argv=None):
     """Setup for main xonsh entry point, returns parsed arguments."""
     if argv is None:
         argv = sys.argv[1:]
+    setup_timings()
     setproctitle = get_setproctitle()
     if setproctitle is not None:
         setproctitle(' '.join(['xonsh'] + argv))
@@ -253,17 +262,11 @@ def premain(argv=None):
     return args
 
 
-def _failback_to_other_shells(argv, err):
-    args = None
-    try:
-        args = premain(argv)
-    except Exception:
-        pass
+def _failback_to_other_shells(args, err):
     # only failback for interactive shell; if we cannot tell, treat it
     # as an interactive one for safe.
     if hasattr(args, 'mode') and args.mode != XonshMode.interactive:
         raise err
-
     foreign_shell = None
     shells_file = '/etc/shells'
     if not os.path.exists(shells_file):
@@ -294,15 +297,16 @@ def _failback_to_other_shells(argv, err):
 
 
 def main(argv=None):
+    args = None
     try:
-        return main_xonsh(argv)
+        args = premain(argv)
+        return main_xonsh(args)
     except Exception as err:
-        _failback_to_other_shells(argv, err)
+        _failback_to_other_shells(args, err)
 
 
-def main_xonsh(argv=None):
+def main_xonsh(args):
     """Main entry point for xonsh cli."""
-    args = premain(argv)
     events.on_post_init.fire()
     env = builtins.__xonsh_env__
     shell = builtins.__xonsh_shell__
