@@ -23,7 +23,8 @@ import threading
 import subprocess
 import collections.abc as cabc
 
-from xonsh.platform import ON_WINDOWS, ON_POSIX, CAN_RESIZE_WINDOW
+from xonsh.platform import (ON_WINDOWS, ON_POSIX, CAN_RESIZE_WINDOW,
+                            LFLAG, CC)
 from xonsh.tools import (redirect_stdout, redirect_stderr, print_exception,
                          XonshCalledProcessError, findfirst, on_main_thread,
                          XonshError, format_std_prepost)
@@ -32,16 +33,6 @@ from xonsh.jobs import wait_for_active_job
 from xonsh.lazyimps import fcntl, termios, _winapi, msvcrt, winutils
 
 
-# termios tc(get|set)attr indexes.
-IFLAG = 0
-OFLAG = 1
-CFLAG = 2
-LFLAG = 3
-ISPEED = 4
-OSPEED = 5
-CC = 6
-
-
 @lazyobject
 def STDOUT_CAPTURE_KINDS():
     return frozenset(['stdout', 'object'])
@@ -758,6 +749,8 @@ class PopenThread(threading.Thread):
         self.send_signal(signum)
         if self.proc is not None and self.proc.poll() is not None:
             self._restore_sigint(frame=frame)
+        if on_main_thread():
+            signal.pthread_kill(threading.get_ident(), signal.SIGINT)
 
     def _restore_sigint(self, frame=None):
         old = self.old_int_handler
@@ -767,7 +760,7 @@ class PopenThread(threading.Thread):
             self.old_int_handler = None
         if frame is not None:
             self._disable_cbreak_stdin()
-            if old is not None:
+            if old is not None and old is not self._signal_int:
                 old(signal.SIGINT, frame)
 
     #
@@ -872,6 +865,16 @@ class PopenThread(threading.Thread):
         """Process return code."""
         return self.proc.returncode
 
+    @property
+    def signal(self):
+        """Process signal, or None."""
+        s = getattr(self.proc, "signal", None)
+        if s is None:
+            rtn = self.returncode
+            if rtn is not None and rtn != 0:
+                s = (-1*rtn, rtn < 0 if ON_WINDOWS else os.WCOREDUMP(rtn))
+        return s
+
     def send_signal(self, signal):
         """Dispatches to Popen.send_signal()."""
         dt = 0.0
@@ -1725,11 +1728,12 @@ class CommandPipeline:
             # we get here if the process is not threadable or the
             # class is the real Popen
             PrevProcCloser(pipeline=self)
-            wait_for_active_job()
-            proc.wait()
-            self._endtime()
-            if self.captured == 'object':
-                self.end(tee_output=False)
+            task = wait_for_active_job()
+            if task is None or task['status'] != 'stopped':
+                proc.wait()
+                self._endtime()
+                if self.captured == 'object':
+                    self.end(tee_output=False)
             raise StopIteration
         # get the correct stderr
         stderr = proc.stderr
@@ -1982,7 +1986,8 @@ class CommandPipeline:
             if core:
                 sig_str += ' (core dumped)'
             print(sig_str, file=sys.stderr)
-            self.errors += sig_str + '\n'
+            if self.errors is not None:
+                self.errors += sig_str + '\n'
 
     def _apply_to_history(self):
         """Applies the results to the current history object."""
@@ -2051,10 +2056,8 @@ class CommandPipeline:
     @property
     def returncode(self):
         """Process return code, waits until command is completed."""
-        proc = self.proc
-        if proc.poll() is None:
-            self.end()
-        return proc.returncode
+        self.end()
+        return self.proc.returncode
 
     rtn = returncode
 
