@@ -13,28 +13,28 @@ are included from the IPython project.  The IPython project is:
 import os
 import sys
 import cmd
-import time
 import select
+import shutil
 import builtins
 import importlib
 import threading
+import subprocess
 import collections
 
-from xonsh.lazyjson import LazyJSON
-from xonsh.lazyasd import LazyObject
+from xonsh.lazyasd import LazyObject, lazyobject
 from xonsh.base_shell import BaseShell
-from xonsh.ansi_colors import ansi_partial_color_format, ansi_color_style_names, ansi_color_style
-from xonsh.prompt.base import partial_format_prompt, multiline_prompt
-from xonsh.tools import print_exception
-from xonsh.platform import ON_WINDOWS, ON_CYGWIN, ON_DARWIN
+from xonsh.ansi_colors import (ansi_partial_color_format, ansi_color_style_names,
+                               ansi_color_style)
+from xonsh.prompt.base import multiline_prompt
+from xonsh.tools import (print_exception, check_for_partial_string, to_bool,
+                         columnize)
+from xonsh.platform import ON_WINDOWS, ON_CYGWIN, ON_DARWIN, ON_POSIX
 from xonsh.lazyimps import pygments, pyghooks
-
-terminal256 = LazyObject(
-    lambda: importlib.import_module('pygments.formatters.terminal256'),
-    globals(), 'terminal')
+from xonsh.events import events
 
 readline = None
 RL_COMPLETION_SUPPRESS_APPEND = RL_LIB = RL_STATE = None
+RL_COMPLETION_QUERY_ITEMS = None
 RL_CAN_RESIZE = False
 RL_DONE = None
 RL_VARIABLE_VALUE = None
@@ -46,7 +46,8 @@ _RL_PREV_CASE_SENSITIVE_COMPLETIONS = 'to-be-set'
 
 def setup_readline():
     """Sets up the readline module and completion suppression, if available."""
-    global RL_COMPLETION_SUPPRESS_APPEND, RL_LIB, RL_CAN_RESIZE, RL_STATE, readline
+    global RL_COMPLETION_SUPPRESS_APPEND, RL_LIB, RL_CAN_RESIZE, RL_STATE, \
+        readline, RL_COMPLETION_QUERY_ITEMS
     if RL_COMPLETION_SUPPRESS_APPEND is not None:
         return
     for _rlmod_name in ('gnureadline', 'readline'):
@@ -78,6 +79,12 @@ def setup_readline():
         except ValueError:
             # not all versions of readline have this symbol, ie Macs sometimes
             RL_COMPLETION_SUPPRESS_APPEND = None
+        try:
+            RL_COMPLETION_QUERY_ITEMS = ctypes.c_int.in_dll(
+                lib, 'rl_completion_query_items')
+        except ValueError:
+            # not all versions of readline have this symbol, ie Macs sometimes
+            RL_COMPLETION_QUERY_ITEMS = None
         try:
             RL_STATE = ctypes.c_int.in_dll(lib, 'rl_readline_state')
         except Exception:
@@ -181,6 +188,18 @@ def rl_completion_suppress_append(val=1):
     RL_COMPLETION_SUPPRESS_APPEND.value = val
 
 
+def rl_completion_query_items(val=None):
+    """Sets the rl_completion_query_items varaiable, if possible.
+    A None value will set this to $COMPLETION_QUERY_LIMIT, otherwise any integer
+    is accepted.
+    """
+    if RL_COMPLETION_QUERY_ITEMS is None:
+        return
+    if val is None:
+        val = builtins.__xonsh_env__.get('COMPLETION_QUERY_LIMIT')
+    RL_COMPLETION_QUERY_ITEMS.value = val
+
+
 def rl_variable_dumper(readable=True):
     """Dumps the currently set readline variables. If readable is True, then this
     output may be used in an inputrc file.
@@ -203,6 +222,17 @@ def rl_variable_value(variable):
     return rtn.decode(encoding=enc, errors=errors)
 
 
+@lazyobject
+def rl_on_new_line():
+    """Grabs one of a few possible redisplay functions in readline."""
+    names = ['rl_on_new_line', 'rl_forced_update_display', 'rl_redisplay']
+    for name in names:
+        func = getattr(RL_LIB, name, None)
+        if func is not None:
+            break
+    return func
+
+
 def _insert_text_func(s, readline):
     """Creates a function to insert text via readline."""
 
@@ -246,7 +276,9 @@ class ReadlineShell(BaseShell, cmd.Cmd):
             except ImportError:
                 store_in_history = True
             pos = readline.get_current_history_length() - 1
+        events.on_pre_prompt.fire()
         rtn = input(self.prompt)
+        events.on_post_prompt.fire()
         if not store_in_history and pos >= 0:
             readline.remove_history_item(pos)
         return rtn
@@ -255,22 +287,55 @@ class ReadlineShell(BaseShell, cmd.Cmd):
         """Overridden to no-op."""
         return '', line, line
 
-    def completedefault(self, text, line, begidx, endidx):
+    def _querycompletions(self, completions, loc):
+        """Returns whether or not we should show completions"""
+        if os.path.commonprefix([c[loc:] for c in completions]):
+            return True
+        elif len(completions) <= builtins.__xonsh_env__.get('COMPLETION_QUERY_LIMIT'):
+            return True
+        msg = '\nDisplay all {} possibilities? '.format(len(completions))
+        msg += '({GREEN}y{NO_COLOR} or {RED}n{NO_COLOR})'
+        self.print_color(msg, end='', flush=True, file=sys.stderr)
+        yn = 'x'
+        while yn not in 'yn':
+            yn = sys.stdin.read(1)
+        show_completions = to_bool(yn)
+        print()
+        if not show_completions:
+            rl_on_new_line()
+            return False
+        w, h = shutil.get_terminal_size()
+        lines = columnize(completions, width=w)
+        more_msg = self.format_color('{YELLOW}==={NO_COLOR} more or '
+                                     '{PURPLE}({NO_COLOR}q{PURPLE}){NO_COLOR}uit '
+                                     '{YELLOW}==={NO_COLOR}')
+        while len(lines) > h - 1:
+            print(''.join(lines[:h-1]), end='', flush=True, file=sys.stderr)
+            lines = lines[h-1:]
+            print(more_msg, end='', flush=True, file=sys.stderr)
+            q = sys.stdin.read(1).lower()
+            print(flush=True, file=sys.stderr)
+            if q == 'q':
+                rl_on_new_line()
+                return False
+        print(''.join(lines), end='', flush=True, file=sys.stderr)
+        rl_on_new_line()
+        return False
+
+    def completedefault(self, prefix, line, begidx, endidx):
         """Implements tab-completion for text."""
+        if self.completer is None:
+            return []
         rl_completion_suppress_append()  # this needs to be called each time
         _rebind_case_sensitive_completions()
-
-        line = builtins.aliases.expand_alias(line)
-        mline = line.partition(' ')[2]
-        offs = len(mline) - len(text)
-        if self.completer is None:
-            x = []
-        else:
-            x = [(i[offs:] if " " in i[:-1] else i)
-                 for i in self.completer.complete(text, line,
-                                                  begidx, endidx,
-                                                  ctx=self.ctx)[0]]
-        return x
+        rl_completion_query_items(val=999999999)
+        completions, l = self.completer.complete(prefix, line,
+                                                 begidx, endidx,
+                                                 ctx=self.ctx)
+        chopped = prefix[:-l]
+        rtn_completions = [chopped + i for i in completions]
+        show_completions = self._querycompletions(completions, endidx - begidx)
+        return rtn_completions if show_completions else []
 
     # tab complete on first index too
     completenames = completedefault
@@ -420,7 +485,7 @@ class ReadlineShell(BaseShell, cmd.Cmd):
         env = builtins.__xonsh_env__  # pylint: disable=no-member
         p = env.get('PROMPT')
         try:
-            p = partial_format_prompt(p)
+            p = self.prompt_formatter(p)
         except Exception:  # pylint: disable=broad-except
             print_exception()
         hide = True if self._force_hide is None else self._force_hide
@@ -430,13 +495,13 @@ class ReadlineShell(BaseShell, cmd.Cmd):
         self.settitle()
         return p
 
-    def format_color(self, string, hide=False, **kwargs):
-        """Readline implementation of color formatting. This usesg ANSI color
+    def format_color(self, string, hide=False, force_string=False, **kwargs):
+        """Readline implementation of color formatting. This usess ANSI color
         codes.
         """
         hide = hide if self._force_hide is None else self._force_hide
-        return ansi_partial_color_format(string, hide=hide,
-                                         style=builtins.__xonsh_env__.get('XONSH_COLOR_STYLE'))
+        style = builtins.__xonsh_env__.get('XONSH_COLOR_STYLE')
+        return ansi_partial_color_format(string, hide=hide, style=style)
 
     def print_color(self, string, hide=False, **kwargs):
         if isinstance(string, str):
@@ -446,7 +511,7 @@ class ReadlineShell(BaseShell, cmd.Cmd):
             env = builtins.__xonsh_env__
             self.styler.style_name = env.get('XONSH_COLOR_STYLE')
             style_proxy = pyghooks.xonsh_style_proxy(self.styler)
-            formatter = terminal256.Terminal256Formatter(style=style_proxy)
+            formatter = pyghooks.XonshTerminal256Formatter(style=style_proxy)
             s = pygments.format(string, formatter).rstrip()
         print(s, **kwargs)
 
@@ -459,11 +524,32 @@ class ReadlineShell(BaseShell, cmd.Cmd):
         style = style = builtins.__xonsh_env__.get('XONSH_COLOR_STYLE')
         return ansi_color_style(style=style)
 
+    def restore_tty_sanity(self):
+        """An interface for resetting the TTY stdin mode. This is highly
+        dependent on the shell backend. Also it is mostly optional since
+        it only affects ^Z backgrounding behaviour.
+        """
+        if not ON_POSIX:
+            return
+        stty, _ = builtins.__xonsh_commands_cache__.lazyget('stty', None)
+        if stty is None:
+            return
+        # If available, we should just call the stty utility. This call should
+        # not throw even if stty fails. It should also be noted that subprocess
+        # calls, like the following, seem to be ineffective:
+        #       subprocess.call([stty, 'sane'], shell=True)
+        # My guess is that this is because Popen does some crazy redirecting
+        # under the covers. This effectively hides the true TTY stdin handle
+        # from stty. To get around this we have to use the lower level
+        # os.system() function.
+        os.system(stty + ' sane')
+
 
 class ReadlineHistoryAdder(threading.Thread):
     def __init__(self, wait_for_gc=True, *args, **kwargs):
-        """Thread responsible for adding inputs from history to the current readline
-        instance. May wait for the history garbage collector to finish.
+        """Thread responsible for adding inputs from history to the
+        current readline instance. May wait for the history garbage
+        collector to finish.
         """
         super(ReadlineHistoryAdder, self).__init__(*args, **kwargs)
         self.daemon = True
@@ -476,22 +562,16 @@ class ReadlineHistoryAdder(threading.Thread):
         except ImportError:
             return
         hist = builtins.__xonsh_history__
-        while self.wait_for_gc and hist.gc.is_alive():
-            time.sleep(0.011)  # gc sleeps for 0.01 secs, sleep a beat longer
-        files = hist.gc.files()
+        if hist is None:
+            return
         i = 1
-        for _, _, f in files:
-            try:
-                lj = LazyJSON(f, reopen=False)
-                for command in lj['cmds']:
-                    inp = command['inp'].splitlines()
-                    for line in inp:
-                        if line == 'EOF':
-                            continue
-                        readline.add_history(line)
-                        if RL_LIB is not None:
-                            RL_LIB.history_set_pos(i)
-                        i += 1
-                lj.close()
-            except (IOError, OSError, ValueError):
+        for h in hist.all_items():
+            line = h['inp'].rstrip()
+            if i == 1:
+                pass
+            elif line == readline.get_history_item(i - 1):
                 continue
+            readline.add_history(line)
+            if RL_LIB is not None:
+                RL_LIB.history_set_pos(i)
+            i += 1
