@@ -1,16 +1,86 @@
 # -*- coding: utf-8 -*-
 """The xonsh shell"""
 import os
+import sys
 import random
+import time
+import difflib
 import builtins
 import warnings
 
-from xonsh.xontribs import update_context
+from xonsh.xontribs import update_context, prompt_xontrib_install
 from xonsh.environ import xonshrc_context
 from xonsh.execer import Execer
 from xonsh.platform import (best_shell_type, has_prompt_toolkit,
                             ptk_version_is_supported)
-from xonsh.tools import XonshError, to_bool_or_int
+from xonsh.tools import XonshError, to_bool_or_int, print_exception
+from xonsh.events import events
+import xonsh.history.main as xhm
+
+
+events.doc('on_transform_command', """
+on_transform_command(cmd: str) -> str
+
+Fired to request xontribs to transform a command line. Return the transformed
+command, or the same command if no transformaiton occurs. Only done for
+interactive sessions.
+
+This may be fired multiple times per command, with other transformers input or
+output, so design any handlers for this carefully.
+""")
+
+events.doc('on_precommand', """
+on_precommand(cmd: str) -> None
+
+Fires just before a command is executed.
+""")
+
+events.doc('on_postcommand', """
+on_postcommand(cmd: str, rtn: int, out: str or None, ts: list) -> None
+
+Fires just after a command is executed. The arguments are the same as history.
+""")
+
+events.doc('on_pre_prompt', """
+on_first_prompt() -> None
+
+Fires just before the prompt is shown
+""")
+
+events.doc('on_post_prompt', """
+on_first_prompt() -> None
+
+Fires just after the prompt returns
+""")
+
+
+def transform_command(src, show_diff=True):
+    """Returns the results of firing the precommand handles."""
+    i = 0
+    limit = sys.getrecursionlimit()
+    lst = ''
+    raw = src
+    while src != lst:
+        lst = src
+        srcs = events.on_transform_command.fire(cmd=src)
+        for s in srcs:
+            if s != lst:
+                src = s
+                break
+        i += 1
+        if i == limit:
+            print_exception('Modifcations to source input took more than '
+                            'the recursion limit number of interations to '
+                            'converge.')
+    debug_level = builtins.__xonsh_env__.get('XONSH_DEBUG')
+    if show_diff and debug_level > 1 and src != raw:
+        sys.stderr.writelines(difflib.unified_diff(
+            raw.splitlines(keepends=True),
+            src.splitlines(keepends=True),
+            fromfile='before precommand event',
+            tofile='after precommand event',
+        ))
+    return src
 
 
 class Shell(object):
@@ -44,6 +114,10 @@ class Shell(object):
                            kwargs.get('scriptcache', True),
                            kwargs.get('cacheall', False))
         env = builtins.__xonsh_env__
+        # build history backend before creating shell
+        builtins.__xonsh_history__ = hist = xhm.construct_history(
+            env=env.detype(), ts=[time.time(), None], locked=True)
+
         # pick a valid shell -- if no shell is specified by the user,
         # shell type is pulled from env
         if shell_type is None:
@@ -79,8 +153,9 @@ class Shell(object):
                              shell_type))
         self.shell = shell_class(execer=self.execer,
                                  ctx=self.ctx, **kwargs)
-        # allows history garbace colector to start running
-        builtins.__xonsh_history__.gc.wait_for_shell = False
+        # allows history garbage colector to start running
+        if hist.gc is not None:
+            hist.gc.wait_for_shell = False
 
     def __getattr__(self, attr):
         """Delegates calls to appropriate shell instance."""
@@ -89,8 +164,10 @@ class Shell(object):
     def _init_environ(self, ctx, config, rc, scriptcache, cacheall):
         self.ctx = {} if ctx is None else ctx
         debug = to_bool_or_int(os.getenv('XONSH_DEBUG', '0'))
+        events.on_timingprobe.fire(name='pre_execer_init')
         self.execer = Execer(config=config, login=self.login, xonsh_ctx=self.ctx,
                              debug_level=debug)
+        events.on_timingprobe.fire(name='post_execer_init')
         self.execer.scriptcache = scriptcache
         self.execer.cacheall = cacheall
         if self.stype != 'none' or self.login:
@@ -98,8 +175,13 @@ class Shell(object):
             names = builtins.__xonsh_config__.get('xontribs', ())
             for name in names:
                 update_context(name, ctx=self.ctx)
+            if getattr(update_context, 'bad_imports', None):
+                prompt_xontrib_install(update_context.bad_imports)
+                del update_context.bad_imports
             # load run control files
             env = builtins.__xonsh_env__
             rc = env.get('XONSHRC') if rc is None else rc
+            events.on_pre_rc.fire()
             self.ctx.update(xonshrc_context(rcfiles=rc, execer=self.execer, initial=self.ctx))
+            events.on_post_rc.fire()
         self.ctx['__name__'] = '__main__'
