@@ -55,26 +55,6 @@ class XonshError(Exception):
     pass
 
 
-class XonshBlockError(XonshError):
-    """Special xonsh exception for communicating the lines of block bodies."""
-
-    def __init__(self, lines, glbs, locs, *args, **kwargs):
-        """
-        Parameters
-        ----------
-        lines : list f str
-            Block lines, as if split by str.splitlines().
-        glbs : Mapping or None
-            Global execution context for lines, ie globals() of calling frame.
-        locs : Mapping or None
-            Local execution context for lines, ie locals() of calling frame.
-        """
-        super().__init__(*args, **kwargs)
-        self.lines = lines
-        self.glbs = glbs
-        self.locs = locs
-
-
 class XonshCalledProcessError(XonshError, subprocess.CalledProcessError):
     """Raised when there's an error with a called process
 
@@ -101,14 +81,33 @@ class XonshCalledProcessError(XonshError, subprocess.CalledProcessError):
         self.completed_command = completed_command
 
 
-def expandpath(path):
-    """Performs environment variable / user expansion on a given path
-    if the relevant flag has been set.
-    """
+def expand_path(s, expand_user=True):
+    """Takes a string path and expands ~ to home if expand_user is set
+    and environment vars if EXPAND_ENV_VARS is set."""
     env = getattr(builtins, '__xonsh_env__', os.environ)
     if env.get('EXPAND_ENV_VARS', False):
-        path = os.path.expanduser(expandvars(path))
-    return path
+        s = expandvars(s)
+    if expand_user:
+        # expand ~ according to Bash unquoted rules "Each variable assignment is
+        # checked for unquoted tilde-prefixes immediately following a ':' or the
+        # first '='". See the following for more details.
+        # https://www.gnu.org/software/bash/manual/html_node/Tilde-Expansion.html
+        pre, char, post = s.partition('=')
+        if char:
+            s = os.path.expanduser(pre) + char
+            s += os.pathsep.join(map(os.path.expanduser, post.split(os.pathsep)))
+        else:
+            s = os.path.expanduser(s)
+    return s
+
+
+def _expandpath(path):
+    """Performs environment variable / user expansion on a given path
+    if EXPAND_ENV_VARS is set.
+    """
+    env = getattr(builtins, '__xonsh_env__', os.environ)
+    expand_user = env.get('EXPAND_ENV_VARS', False)
+    return expand_path(path, expand_user=expand_user)
 
 
 def decode_bytes(b):
@@ -172,9 +171,9 @@ class EnvPath(collections.MutableSequence):
     def __getitem__(self, item):
         # handle slices separately
         if isinstance(item, slice):
-            return [expandpath(i) for i in self._l[item]]
+            return [_expandpath(i) for i in self._l[item]]
         else:
-            return expandpath(self._l[item])
+            return _expandpath(self._l[item])
 
     def __setitem__(self, index, item):
         self._l.__setitem__(index, item)
@@ -347,17 +346,25 @@ def subproc_toks(line, mincol=-1, maxcol=None, lexer=None, returnline=False):
     return rtn
 
 
+def is_balanced(expr, ltok, rtok):
+    """Determines whether an expression has unbalanced opening and closing tokens."""
+    lcnt = expr.count(ltok)
+    if lcnt == 0:
+        return True
+    rcnt = expr.count(rtok)
+    if lcnt == rcnt:
+        return True
+    else:
+        return False
+
+
 def subexpr_from_unbalanced(expr, ltok, rtok):
     """Attempts to pull out a valid subexpression for unbalanced grouping,
     based on opening tokens, eg. '(', and closing tokens, eg. ')'.  This
     does not do full tokenization, but should be good enough for tab
     completion.
     """
-    lcnt = expr.count(ltok)
-    if lcnt == 0:
-        return expr
-    rcnt = expr.count(rtok)
-    if lcnt == rcnt:
+    if is_balanced(expr, ltok, rtok):
         return expr
     subexpr = expr.rsplit(ltok, 1)[-1]
     subexpr = subexpr.rsplit(',', 1)[-1]
@@ -365,6 +372,19 @@ def subexpr_from_unbalanced(expr, ltok, rtok):
     return subexpr
 
 
+def subexpr_before_unbalanced(expr, ltok, rtok):
+    """Obtains the expression prior to last unblanced left token."""
+    subexpr, _, post = expr.rpartition(ltok)
+    nrtoks_in_post = post.count(rtok)
+    while nrtoks_in_post != 0:
+        for i in range(nrtoks_in_post):
+            subexpr, _, post = subexpr.rpartition(ltok)
+        nrtoks_in_post = post.count(rtok)
+    _, _, subexpr = subexpr.rpartition(rtok)
+    _, _, subexpr = subexpr.rpartition(ltok)
+    return subexpr
+
+
 def decode(s, encoding=None):
     encoding = encoding or DEFAULT_ENCODING
     return s.decode(encoding, "replace")
@@ -673,6 +693,7 @@ def is_writable_file(filepath):
     """
     Checks if a filepath is valid for writing.
     """
+    filepath = expand_path(filepath)
     # convert to absolute path if needed
     if not os.path.isabs(filepath):
         filepath = os.path.abspath(filepath)
@@ -819,6 +840,11 @@ def is_string_or_callable(x):
     return is_string(x) or is_callable(x)
 
 
+def is_class(x):
+    """Tests if something is a class"""
+    return isinstance(x, type)
+
+
 def always_true(x):
     """Returns True"""
     return True
@@ -915,6 +941,11 @@ def to_bool(x):
         return bool(x)
 
 
+def to_itself(x):
+    """No conversion, returns itself."""
+    return x
+
+
 def bool_to_str(x):
     """Converts a bool to an empty string if False and the string '1' if
     True.
@@ -1240,6 +1271,11 @@ def is_history_tuple(x):
     return False
 
 
+def is_history_backend(x):
+    """Tests if something is a valid history backend."""
+    return is_string(x) or is_class(x) or isinstance(x, object)
+
+
 def is_dynamic_cwd_width(x):
     """ Determine if the input is a valid input for the DYNAMIC_CWD_WIDTH
     environement variable.
@@ -1701,3 +1737,39 @@ def format_datetime(dt):
     """Format datetime object to string base on $XONSH_DATETIME_FORMAT Env."""
     format_ = builtins.__xonsh_env__['XONSH_DATETIME_FORMAT']
     return dt.strftime(format_)
+
+
+def columnize(elems, width=80, newline='\n'):
+    """Takes an iterable of strings and returns a list of lines with the
+    elements placed in columns. Each line will be at most *width* columns.
+    The newline character will be appended to the end of each line.
+    """
+    sizes = [len(e) + 1 for e in elems]
+    nelem = len(elems)
+    ncols = 1
+    nrows = len(sizes)
+    columns = [sizes]
+    last_longest_row = max(sizes)
+    while True:
+        longest_row = sum(map(max, columns))
+        if longest_row - 1 <= width:
+            # we might be able to fit another column.
+            ncols += 1
+            nrows = nelem // ncols
+            columns = [sizes[i*nrows:(i+1)*nrows] for i in range(ncols)]
+            last_longest_row = longest_row
+        else:
+            # we can't fit another column
+            ncols -= 1
+            nrows = nelem // ncols
+            break
+    pad = (width - last_longest_row + ncols) // ncols
+    pad = pad if pad > 1 else 1
+    data = [elems[i*nrows:(i+1)*nrows] for i in range(ncols)]
+    colwidths = [max(map(len, d)) + pad for d in data]
+    colwidths[-1] -= pad
+    row_t = ''.join(['{{row[{i}]: <{{w[{i}]}}}}'.format(i=i) for i in range(ncols)])
+    row_t += newline
+    lines = [row_t.format(row=row, w=colwidths) for row in
+             itertools.zip_longest(*data, fillvalue='')]
+    return lines
